// File_MpegPs - Info for MPEG files
// Copyright (C) 2002-2007 Jerome Martinez, Zen@MediaArea.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//---------------------------------------------------------------------------
// Compilation conditions
#include <MediaInfo/Setup.h>
#ifdef __BORLANDC__
    #pragma hdrstop
#endif
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#if defined(MEDIAINFO_MPEGPS_YES)
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "MediaInfo/Multiple/File_MpegPs.h"
#if defined(MEDIAINFO_AVC_YES)
    #include "MediaInfo/Video/File_Avc.h"
#endif
#if defined(MEDIAINFO_MPEG4V_YES)
    #include "MediaInfo/Video/File_Mpeg4v.h"
#endif
#if defined(MEDIAINFO_MPEGV_YES)
    #include "MediaInfo/Video/File_Mpegv.h"
#endif
#if defined(MEDIAINFO_VC1_YES)
    #include "MediaInfo/Video/File_Vc1.h"
#endif
#if defined(MEDIAINFO_AC3_YES)
    #include "MediaInfo/Audio/File_Ac3.h"
#endif
#if defined(MEDIAINFO_DTS_YES)
    #include "MediaInfo/Audio/File_Dts.h"
#endif
#if defined(MEDIAINFO_MPEGA_YES)
    #include "MediaInfo/Audio/File_Mpega.h"
#endif
#if defined(MEDIAINFO_PCM_YES)
    #include "MediaInfo/Audio/File_Pcm.h"
#endif
#if defined(MEDIAINFO_RLE_YES)
    #include "MediaInfo/Image/File_Rle.h"
#endif
#include <ZenLib/Utils.h>
using namespace ZenLib;
//---------------------------------------------------------------------------

namespace MediaInfoLib
{

//***************************************************************************
// Constants
//***************************************************************************

//---------------------------------------------------------------------------
const char* MpegPs_Version[]=
{
    "MPEG-1PS",
    "MPEG-2PS",
    "???",
    "???",
};

//---------------------------------------------------------------------------
const char* MpegPs_System_Fixed[]=
{
    "CBR",
    "VBR",
};

//---------------------------------------------------------------------------
const char* MpegPs_StreamType(int16u Element_Name)
{
         if (Element_Name>=0xC0
          && Element_Name<=0xDF) return "MPEG Audio";
    else if (Element_Name>=0xE0
          && Element_Name<=0xEF) return "MPEG Video";
    else if (Element_Name==0xB8) return "For all MPEG Audio streams";
    else if (Element_Name==0xB9) return "For all MPEG Video streams";
    else if (Element_Name==0xBD) return "Private 1";
    else if (Element_Name==0xBF) return "Private 2";
    else if (Element_Name==0xFD) return "Private HD";
    else                         return "";
};

//---------------------------------------------------------------------------
extern const char* MpegTs_Pms_StreamID(int32u ID);

//***************************************************************************
// Format
//***************************************************************************

//---------------------------------------------------------------------------
File_MpegPs::~File_MpegPs()
{
    for (size_t Stream_Pos=0; Stream_Pos<Video_Handle.size(); Stream_Pos++)
        delete Video_Handle[Stream_Pos]; //Video_Handle[Stream_Pos]=NULL;
    for (size_t Stream_Pos=0; Stream_Pos<VideoMpeg4v_Handle.size(); Stream_Pos++)
        delete VideoMpeg4v_Handle[Stream_Pos]; //Video_Handle[Stream_Pos]=NULL;
    for (size_t Stream_Pos=0; Stream_Pos<VideoAvc_Handle.size(); Stream_Pos++)
        delete VideoAvc_Handle[Stream_Pos]; //Video_Handle[Stream_Pos]=NULL;
    for (size_t Stream_Pos=0; Stream_Pos<Audio_Handle.size(); Stream_Pos++)
        delete Audio_Handle[Stream_Pos]; //Audio_Handle[Stream_Pos]=NULL;
    for (size_t Stream_Pos=0; Stream_Pos<Private1_Handle.size(); Stream_Pos++)
        delete Private1_Handle[Stream_Pos]; //Private1_Handle[Stream_Pos]=NULL;
}

//---------------------------------------------------------------------------
void File_MpegPs::Read_Buffer_Init()
{
    MPEG_Version=0;
    Video_Unlimited=false;
    Element_Name=0;
    Element_Size=0;
    Element_HeaderSize=0;
    Buffer_MinimumSize=100;
    TimeStamp.PTS=0;
    TimeStamp.DTS=0;
    FromTS=0; //No info
    Synched=false;
    Video_Count=(int8u)-1;
    Audio_Count=(int8u)-1;
}

//---------------------------------------------------------------------------
void File_MpegPs::Read_Buffer_Unsynched()
{
    if (Config.Details_Get())
    {
        Details_Add_Error("Will resync...");
    }

    Synched=false;
    Video_Unlimited=false;
    Video_Count=(int8u)-1;
    Audio_Count=(int8u)-1;
}

//---------------------------------------------------------------------------
void File_MpegPs::Read_Buffer_Continue()
{
    //Integrity
    if (File_Offset==0 && Detect_NonMPEGPS())
    {
        File_Offset=File_Size;
        return;
    }

    //Video unlimited specific, we didn't wait for the end (because this is... unlimited)
    if (Video_Unlimited)
    {
        Element_Size_Get_Unlimited();
        Element_Parse();
        Buffer_Offset+=Element_Size;
    }

    //Buffer
    Element_Next=File_Offset;

    //Parse with PES parser
    while (Buffer_Parse());
}

//---------------------------------------------------------------------------
void File_MpegPs::Read_Buffer_Finalize()
{
    //Video
    Video.clear();
    for (size_t Stream_Pos=0; Stream_Pos<Video_Handle.size(); Stream_Pos++)
        if (Video_Handle[Stream_Pos])
        {

            //If no success
            if (Video_Handle[Stream_Pos]->Count_Get(Stream_Video)==0)
            {
                Stream_Prepare(Stream_Video);
            }
            else
            {
                Open_Buffer_Finalize(Video_Handle[Stream_Pos]);
                if (Merge(*Video_Handle[Stream_Pos])>0)
                {
                    //Delay
                    if (Video_TimeStamp[Stream_Pos].Have_PTS)
                    {
                        Fill("Delay", "");
                        Fill("Delay", Video_TimeStamp[Stream_Pos].PTS/90);
                    }

                    //TimeStamp
                    int64s PlayTime=Video_TimeStamp_End[Stream_Pos].PTS-Video_TimeStamp[Stream_Pos].PTS;
                    if (PlayTime && Get(Stream_Video, StreamPos_Last, _T("PlayTime")).empty())
                        Fill("PlayTime", PlayTime/90);

                    //Tags in MPEG Video
                    Fill(Stream_General, 0, "Encoded_Library", Get(Stream_Video, StreamPos_Last, _T("Encoded_Library")));
                }
            }

            delete Video_Handle[Stream_Pos]; Video_Handle[Stream_Pos]=NULL;
        }

    //Audio
    Audio.clear();
    for (size_t Stream_Pos=0; Stream_Pos<Audio_Handle.size(); Stream_Pos++)
        if (Audio_Handle[Stream_Pos])
        {

            Open_Buffer_Finalize(Audio_Handle[Stream_Pos]);
            if (Merge(*Audio_Handle[Stream_Pos])>0)
            {
                //Delay
                if (Audio_TimeStamp[Stream_Pos].Have_PTS)
                {
                    Fill("Delay", "");
                    Fill("Delay", (Audio_TimeStamp[Stream_Pos].PTS)/90);
                }

                //TimeStamp
                int64s PlayTime=Audio_TimeStamp_End[Stream_Pos].PTS-Audio_TimeStamp[Stream_Pos].PTS;
                if (PlayTime && Get(Stream_Audio, StreamPos_Last, _T("PlayTime")).empty())
                    Fill("PlayTime", PlayTime/90);
            }

            delete Audio_Handle[Stream_Pos]; Audio_Handle[Stream_Pos]=NULL;
        }

    //Private1
    for (size_t Stream_Pos=0; Stream_Pos<Private1_Handle.size(); Stream_Pos++)
        if (Private1_Handle[Stream_Pos])
        {

            Open_Buffer_Finalize(Private1_Handle[Stream_Pos]);
            if (Merge(*Private1_Handle[Stream_Pos])>0)
            {
                //Delay
                if (Private1_TimeStamp[Stream_Pos].Have_PTS)
                {
                    Fill("Delay", "");
                    Fill("Delay", Private1_TimeStamp[Stream_Pos].PTS/90);
                }

                //TimeStamp
                int64s PlayTime=Private1_TimeStamp_End[Stream_Pos].PTS-Private1_TimeStamp[Stream_Pos].PTS;
                if (PlayTime && Get(Stream_Audio, StreamPos_Last, _T("PlayTime")).empty())
                    Fill("PlayTime", PlayTime/90);
            }

            delete Private1_Handle[Stream_Pos]; Private1_Handle[Stream_Pos]=NULL;
        }

    //Fill General
    if (General.empty())
        Stream_Prepare(Stream_General);

         if (!Video.empty() && Video[0](_T("Codec"))==_T("MPEG-2V"))
        Fill(Stream_General, 0, "Format", "MPEG-2PS");
    else if (!Video.empty() && Video[0](_T("Codec"))==_T("MPEG-4V"))
        Fill(Stream_General, 0, "Format", "MPEG-4PS");
    else if (MPEG_Version==2)
        Fill(Stream_General, 0, "Format", "MPEG-2PS");
    else if (MPEG_Version==1)
        Fill(Stream_General, 0, "Format", "MPEG-1PS");
}

//***************************************************************************
// Buffer
//***************************************************************************

//---------------------------------------------------------------------------
bool File_MpegPs::Buffer_Parse()
{
    //Next Frame
    if (!NextFrame())
        return false;

    //Element name
    if (!Element_Name_Get())
        return false;

    //Element size
    if (!Element_Size_Get())
        return false;

    //Is OK?
    if (Element_Name_IsOK() && !Element_Parse())
        return false;

    //Jumping to the end of the file if needed
    if (Detect_EOF_Needed())
        return false;

    //If no need of more
    if (File_Offset==File_Size || File_GoTo!=(int64u)-1)
        return false;

    Buffer_Offset=Element_Next;
    return true;
}

//---------------------------------------------------------------------------
//
bool File_MpegPs::Element_Parse()
{
    ELEMENT(0, Element_Name, Element_Size);

    //Positionning
    Buffer_Offset+=Element_HeaderSize;

    //Parsing
    switch (Element_Name)
    {
        case 0xB9 : MPEG_program_end(); break;
        case 0xBA : pack_start(); break;
        case 0xBB : system_header_start(); break;
        case 0xBC : program_stream_map(); break;
        case 0xBD : private_stream_1(); break;
        case 0xBE : padding_stream(); break;
        case 0xBF : private_stream_2(); break;
        case 0xFD : if (FromTS)
                        private_stream_1();
                    else
                        video_stream();
                    break;
        default:
                 if (Element_Name>=0xC0
                  && Element_Name<=0xDF) audio_stream();
            else if (Element_Name>=0xE0
                  && Element_Name<=0xEF) video_stream();
            else
                TRUSTED_ISNOT("Unattended element!");
    }

    FLUSH();
    return true;
}

//***************************************************************************
// Elements
//***************************************************************************

//---------------------------------------------------------------------------
// Packet "B9"
void File_MpegPs::MPEG_program_end()
{
    NAME("MPEG_program_end");

    //Parsing
    CC_BEGIN();
    CC_END();
}

//---------------------------------------------------------------------------
// Packet "BA"
void File_MpegPs::pack_start()
{
    NAME("pack_start");
    NOT_NEEDED(Count_Get(Stream_General)>0);

    //Reading
    int32u Version, SysClock32, SysClock29, SysClock14, Padding;
    BS_BEGIN()
    GET_BS ( 2, Version,                                        version) INFO(MpegPs_Version[Version]);
    if (Version==1)
    {
        //MPEG-2
        MPEG_Version=2;
        GET_BS ( 3, SysClock32,                                 system_clock_reference_base32)
        MARK_1(                                                 marker_bit)
        GET_BS (15, SysClock29,                                 system_clock_reference_base29)
        MARK_1(                                                 marker_bit)
        GET_BS (15, SysClock14,                                 system_clock_reference_base14)
        MARK_1(                                                 marker_bit)
        SKIP_BS( 9,                                             system_clock_reference_extension)
        MARK_1(                                                 marker_bit)
        SKIP_BS(22,                                             program_mux_rate)
        MARK_1(                                                 marker_bit)
        MARK_1(                                                 marker_bit)
        SKIP_BS( 5,                                             reserved)
        GET_BS ( 3, Padding,                                    pack_stuffing_length)
        SKIP_BS(8*Padding,                                      padding)
        BS_END();

        //int64u system_clock_reference_base=SysClock32<<30
        //                                || SysClock29<<15
        //                                || SysClock14<<00
        //                                ;
    }
    else
    {
        //MPEG-1
        MPEG_Version=1;
        SKIP_BS( 1,                                             sync)
        SKIP_BS( 1,                                             sync)
        GET_BS ( 3, SysClock32,                                 system_clock_reference_base32)
        MARK_1(                                                 marker_bit)
        GET_BS (15, SysClock29,                                 system_clock_reference_base29)
        MARK_1(                                                 marker_bit)
        GET_BS (15, SysClock14,                                 system_clock_reference_base14)
        MARK_1(                                                 marker_bit)
        MARK_1(                                                 marker_bit)
        SKIP_BS(22,                                             program_mux_rate)
        MARK_1(                                                 marker_bit)
        BS_END()
    }

    //Filling
    if (Count_Get(Stream_General)==0)
        Stream_Prepare(Stream_General);
}

//---------------------------------------------------------------------------
// Packet "BB"
void File_MpegPs::system_header_start()
{
    NAME("system_header_start");
    INTEGRITY_GENERAL(0);

    int32u Audio_Count_Temp, Video_Count_Temp;
    BS_BEGIN()
    MARK_1(                                                     marker_bit)
    SKIP_BS(22,                                                 rate_bound)
    MARK_1(                                                     marker_bit)
    GET_BS ( 6, Audio_Count_Temp,                               audio_bound)
    INFO_BS( 1, fixed_flag,                                     fixed_flag); INFO(MpegPs_System_Fixed[fixed_flag]);
    SKIP_BS( 1,                                                 CSPS_flag)
    SKIP_BS( 1,                                                 system_audio_lock_flag)
    SKIP_BS( 1,                                                 system_video_lock_flag)
    MARK_1(                                                     marker_bit)
    GET_BS ( 5, Video_Count_Temp,                               video_bound)
    SKIP_BS( 1,                                                 packet_rate_restriction_flag)
    SKIP_BS( 7,                                                 reserved_byte)
    while (BS.Peek(1)==1)
    {
        FLUSH();
        ELEMENT_BEGIN();
        int32u stream_id, Scale, Size;
        GET_BS ( 8, stream_id,                                  stream_id)
        ELEMENT(1, (int16u)stream_id, 3);
        NAME(MpegPs_StreamType((int16u)stream_id));
        INFO(MpegPs_StreamType((int16u)stream_id));
        MARK_1(                                                 marker_bit)
        MARK_1(                                                 marker_bit)
        GET_BS ( 1, Scale,                                      STD_buffer_bound_scale)
        GET_BS (13, Size,                                       STD_buffer_size_bound) INFO(Ztring::ToZtring(Size*(Scale?1024:128)) + _T(" bytes"));
        ELEMENT_END();
    }
    BS_END();

    //Filling
    if (Audio_Count==(int8u)-1) //0xBB may be multipart
        Audio_Count=0;
    Audio_Count+=Audio_Count_Temp;
    if (Video_Count==(int8u)-1) //0xBB may be multipart
        Video_Count=0;
    Video_Count+=Video_Count_Temp;
}

//---------------------------------------------------------------------------
// Packet "BC"
void File_MpegPs::program_stream_map()
{
    NAME("program_stream_map");
    MPEG_Version=2;
    StreamKind.resize(0x100); //Init

    int32u PSI_Size;
    BS_BEGIN()
    SKIP_BS( 1,                                                 current_next_indicator)
    SKIP_BS( 2,                                                 reserved)
    SKIP_BS( 5,                                                 program_stream_map_version)
    SKIP_BS( 7,                                                 reserved)
    MARK_1(                                                     marker_bit)
    GET_BS (16, PSI_Size,                                       program_stream_info_length)
    for (int32u Descriptor_Pos=0; Descriptor_Pos<PSI_Size; Descriptor_Pos++)
    {
        //Descriptor
        //SKIP_BS( 8,                                             descriptor)
        BS.Skip(8);
    }
    int32u ESM_Size, ES_Size, ESID, StreamID;
    size_t Streams_Count_Temp=0;
    int32u Pos=0;
    GET_BS (16, ESM_Size,                                       elementary_stream_map_length)
    while (Pos<ESM_Size)
    {
        GET_BS ( 8, StreamID,                                   stream_type) INFO(MpegTs_Pms_StreamID(StreamID));
        GET_BS ( 8, ESID,                                       elementary_stream_id)
        GET_BS (16, ES_Size,                                    ES_info_length)
        for (int32u Descriptor_Pos=0; Descriptor_Pos<ES_Size; Descriptor_Pos++)
        {
            //Descriptor
            //SKIP_BS( 8,                                            descriptor)
            BS.Skip(8);
        }

        StreamKind[ESID]=StreamID;
        Pos+=4+ES_Size;
        Streams_Count_Temp++;
    }
    SKIP_BS(32,                                                 CRC_32)
    BS_END();
}

//---------------------------------------------------------------------------
// Packet "BD"
void File_MpegPs::private_stream_1()
{
    NAME("private_stream_1");

    //For TS streams in debug mode, which does not have Start chunk
    if (General.empty() && FromTS>0 && (Element_Size>10000 || Buffer_Offset+Element_Size+10000>=Buffer_Size)) //Trying to detect false positive
    {
        Stream_Prepare(Stream_General);
        Stream_Prepare(Stream_Audio);
        Video_Count=0;
        Audio_Count=1;
    }

    //Testing DVD
    if (!FromTS)
    {
        Private1_ChooseID();
        if (Private1_ID>=Private1_Handle.size() || Private1_Handle[Private1_ID]==NULL || Private1_Handle[Private1_ID]->Count_Get(Stream_General)==0)
        {
            FLUSH();
            ELEMENT(1, "Testing DVD Private...", Element_Size);
        }
    }
    else
    {
        Private1_ID=0;
        Private1_Offset=0;
    }
    
    //New stream if needed
    if (Private1_ID>=Private1_Handle.size())
    {
        Private1_Handle.resize(Private1_ID+1);
        Private1_TimeStamp.resize(Private1_ID+1);
        Private1_TimeStamp_End.resize(Private1_ID+1);
    }
    if (Private1_Handle[Private1_ID]==NULL)
    {
        Private1_Handle[Private1_ID]=Private1_ChooseParser();
        if (Private1_Handle[Private1_ID]==NULL)
            return;
        Private1_TimeStamp[Private1_ID]=TimeStamp;
    }

    //TimeStamp
    Private1_TimeStamp_End[Private1_ID]=TimeStamp;

    //Open
    Open_Buffer_Init(Private1_Handle[Private1_ID], File_Size, File_Offset+Buffer_Offset+Private1_Offset);
    Open_Buffer_Continue(Private1_Handle[Private1_ID], Buffer+Buffer_Offset+Private1_Offset, Element_Size-Private1_Offset);

    //Testing other parsers in case of need
    if (StreamKind.empty() && Private1_Handle[Private1_ID]->Count_Get(Stream_General)==0)
    {
        FLUSH();
        ELEMENT(1, "Testing AC3...", Element_Size);
        if (Private1_ID>=Private1Ac3_Handle.size())
            Private1Ac3_Handle.resize(Private1_ID+1);
        if (Private1Ac3_Handle[Private1_ID]==NULL)
            Private1Ac3_Handle[Private1_ID]=new File_Ac3;
        Open_Buffer_Init(Private1Ac3_Handle[Private1_ID], File_Size, File_Offset+Buffer_Offset);
        Open_Buffer_Continue(Private1Ac3_Handle[Private1_ID], Buffer+Buffer_Offset, Element_Size);

        if (Private1Ac3_Handle[Private1_ID]->Count_Get(Stream_General)>0)
        {
            ELEMENT(1, "Ac3 found, changing default parser", 0);
            delete Private1_Handle[Private1_ID]; Private1_Handle[Private1_ID]=Private1Ac3_Handle[Private1_ID]; Private1Ac3_Handle[Private1_ID]=NULL;
        }
        else
        {
            FLUSH();
            ELEMENT(1, "Testing DTS...", Element_Size);
            if (Private1_ID>=Private1Dts_Handle.size())
                Private1Dts_Handle.resize(Private1_ID+1);
            if (Private1Dts_Handle[Private1_ID]==NULL)
                Private1Dts_Handle[Private1_ID]=new File_Dts;
            Open_Buffer_Init(Private1Dts_Handle[Private1_ID], File_Size, File_Offset+Buffer_Offset);
            Open_Buffer_Continue(Private1Dts_Handle[Private1_ID], Buffer+Buffer_Offset, Element_Size);

            if (Private1Dts_Handle[Private1_ID]->Count_Get(Stream_General)>0)
            {
                ELEMENT(1, "DTS found, changing default parser", 0);
                delete Private1_Handle[Private1_ID]; Private1_Handle[Private1_ID]=Private1Dts_Handle[Private1_ID]; Private1Dts_Handle[Private1_ID]=NULL;
            }
        }
    }

    if (Private1_ID<0x20 || Private1_ID>0x3F)
        if (Private1_Handle[Private1_ID] && Private1_Handle[Private1_ID]->Count_Get(Stream_Audio)>0 && Audio_Count>0 && (Private1_Handle[Private1_ID]->File_GoTo!=(int64u)-1 || Private1_Handle[Private1_ID]->File_Offset==Private1_Handle[Private1_ID]->File_Size))
            Audio_Count--;
}

//---------------------------------------------------------------------------
void File_MpegPs::Private1_ChooseID()
{
    //Init
    Private1_Offset=0;

    //From TS
    if (FromTS>0)
    {
        Private1_ID=FromTS;
        return;
    }

    //Trying VOB system
    Private1_ID=0;
    Private1_ChooseID_VOB();
    if (Private1_ID)
        return;

    /*
    //Unknown?
    else
    {
        //Managing Transport Streams with ID from TS but with first byte indicating kind.
        if (Element_Size<1)
            return NULL;
        int8u Private1_ID=Buffer[Buffer_Offset];
        Buffer_Offset++;
        Element_Size--;
        return Private1_ChooseParser(Private1_ID);
    }
    */
}

//---------------------------------------------------------------------------
void File_MpegPs::Private1_ChooseID_VOB()
{
    //Parsing
    int32u CodecID, Count, Next;
    CC_BEGIN();
    GET_B1 (CodecID,                                        CodecID);
    GET_B1 (Count,                                          Count of next frame headers);
    GET_B2 (Next,                                           Next frame offset minus 1);
    CC_END_CANBEMORE();

    //Coherancy
    if (Next+1+4>Element_Size)
        return; //Should not!

    //Testing
    if (Count>0)
    {
        //Subtitles
             if (CodecID>=0x20 && CodecID<=0x3F)
            ;
        //AC3
        else if (CodecID>=0x80 && CodecID<=0x87)
        {
            if (CC2(Buffer+Buffer_Offset+3+Next)!=0x0B77)
                return;
        }
        //DTS
        else if (CodecID>=0x88 && CodecID<=0x8F)
        {
            if (CC4(Buffer+Buffer_Offset+3+Next)!=0x7FFE8001)
                return;
        }
        //PCM
        else if (CodecID>=0xA0 && CodecID<=0xA8)
            return ;
        //AC3+
        else if (CodecID>=0xC0 && CodecID<=0xC7)
        {
            if (CC2(Buffer+Buffer_Offset+3+Next)!=0x0B77)
                return;
        }
        //Unknown
        else
        {
            Private1_ID=0;
            return;
        }
    }
    
    //Filling
    Private1_ID=(int8u)CodecID;

    //Positionning
    Private1_Offset=4;
}

//---------------------------------------------------------------------------
File__Base* File_MpegPs::Private1_ChooseParser()
{
    if (Private1_ID)
        NAME(Ztring::ToZtring(Private1_ID, 16));

    //Subtitles
         if (Private1_ID>=0x20 && Private1_ID<=0x3F)
        return Private1_ChooseParser_RLE();
    //AC3
    else if (Private1_ID>=0x80 && Private1_ID<=0x87)
        return Private1_ChooseParser_AC3();
    //DTS
    else if (Private1_ID>=0x88 && Private1_ID<=0x8F)
        return Private1_ChooseParser_DTS();
    //PCM
    else if (Private1_ID>=0xA0 && Private1_ID<=0xA8)
        return Private1_ChooseParser_PCM();
    //AC3+
    else if (Private1_ID>=0xC0 && Private1_ID<=0xC7)
        return Private1_ChooseParser_AC3();
    else
        return Private1_ChooseParser_NULL();
}

//---------------------------------------------------------------------------
File__Base* File_MpegPs::Private1_ChooseParser_RLE()
{
    NAME("RLE");

    //Filling
    #if defined(MEDIAINFO_RLE_YES)
        return new File_Rle();
    #else
        //Filling
        File__Base* Handle=new File__Base();
        Handle->Stream_Prepare(Stream_Text);
        Handle->Fill("Codec", "RLE");
        return Handle;
    #endif
}

//---------------------------------------------------------------------------
File__Base* File_MpegPs::Private1_ChooseParser_AC3()
{
    NAME("AC3");

    //Filling
    #if defined(MEDIAINFO_AC3_YES)
        return new File_Ac3();
    #else
        //Filling
        File__Base* Handle=new File__Base();
        Handle->Stream_Prepare(Stream_Audio);
        Handle->Fill("Codec", "AC3");
        return Handle;
    #endif
}

//---------------------------------------------------------------------------
File__Base* File_MpegPs::Private1_ChooseParser_DTS()
{
    NAME("DTS");

    //Filling
    #if defined(MEDIAINFO_DTS_YES)
        return new File_Dts();
    #else
        //Filling
        File__Base* Handle=new File__Base();
        Handle->Stream_Prepare(Stream_Audio);
        Handle->Fill("Codec", "DTS");
        return Handle;
    #endif
}

//---------------------------------------------------------------------------
File__Base* File_MpegPs::Private1_ChooseParser_PCM()
{
    NAME("PCM");

    //Filling
    #if defined(MEDIAINFO_PCM_YES)
        return new File_Pcm();
    #else
        //Filling
        File__Base* Handle=new File__Base();
        Handle->Stream_Prepare(Stream_Audio);
        Handle->Fill("Codec", "PCM");
        return Handle;
    #endif
}

//---------------------------------------------------------------------------
File__Base* File_MpegPs::Private1_ChooseParser_NULL()
{
    //Filling
    File__Base* Handle=new File__Base();
    return Handle;
}

//---------------------------------------------------------------------------
// Packet "BE"
void File_MpegPs::padding_stream()
{
    NAME("padding_stream");
}

//---------------------------------------------------------------------------
// Packet "BF"
void File_MpegPs::private_stream_2()
{
    NAME("private_stream_2");
}

//---------------------------------------------------------------------------
void File_MpegPs::audio_stream()
{
    NAME("Audio");

    //For TS streams in debug mode, which does not have Start chunk
    if (FromTS>0 && General.empty() && (Element_Size>10000 || Buffer_Offset+Element_Size+10000>=Buffer_Size)) //Trying to detect false positive
    {
        Stream_Prepare(Stream_General);
        Stream_Prepare(Stream_Audio);
        Video_Count=0;
        Audio_Count=1;
        StreamKind.resize(0x100);
        StreamKind[Element_Name]=FromTS;
    }

    //New stream if needed
    int8u Stream_Pos=Element_Name-0xC0;
    if (Stream_Pos>=Audio_Handle.size())
    {
        Audio_Handle.resize(Stream_Pos+1);
        Audio_TimeStamp.resize(Stream_Pos+1);
        Audio_TimeStamp_End.resize(Stream_Pos+1);
    }
    if (Audio_Handle[Stream_Pos]==NULL)
    {
             if (StreamKind.empty() || StreamKind[Element_Name]==0x03 || StreamKind[Element_Name]==0x04)
        {
            #if defined(MEDIAINFO_MPEGA_YES)
                Audio_Handle[Stream_Pos]=new File_Mpega;
                ((File_Mpega*)Audio_Handle[Stream_Pos])->Frame_Count_Valid=1;
            #else
                //Filling
                Audio_Handle[Stream_Pos]=new File__Base();
                Audio_Handle[Stream_Pos]->Stream_Prepare(Stream_Audio);
                Audio_Handle[Stream_Pos]->Fill("Codec", "MPEG Audio");
            #endif
        }
        else
        {
            Audio_Handle[Stream_Pos]=new File__Base();
            Audio_Handle[Stream_Pos]->Stream_Prepare(Stream_Video);
            Audio_Handle[Stream_Pos]->Fill("Codec", "Unsupported");
        }

        Audio_TimeStamp[Stream_Pos]=TimeStamp;
    }

    //TimeStamp
    Audio_TimeStamp_End[Stream_Pos]=TimeStamp;

    //Open Mpega
    Open_Buffer_Init(Audio_Handle[Stream_Pos], File_Size, File_Offset+Buffer_Offset);
    Open_Buffer_Continue(Audio_Handle[Stream_Pos], Buffer+Buffer_Offset, Element_Size);
    if (Audio_Count>0 && (Audio_Handle[Stream_Pos]->File_GoTo!=(int64u)-1 || Audio_Handle[Stream_Pos]->File_Offset==Audio_Handle[Stream_Pos]->File_Size))
        Audio_Count--;

    //Demux
    DEMUX(Buffer+Buffer_Offset, Element_Size, Ztring::ToZtring(Element_Name, 16)+_T(".mpa"));
}

//---------------------------------------------------------------------------
void File_MpegPs::video_stream()
{
    NAME ("Video");

    if (!File_Name.empty())
    {
        //INTEGRITY_GENERAL()
    }

    //For TS streams in debug mode, which does not have Start chunk
    if (FromTS>0 && General.empty() && (Element_Size>10000 || Buffer_Offset+Element_Size+10000>=Buffer_Size)) //Trying to detect false positive
    {
        Stream_Prepare(Stream_General);
        Stream_Prepare(Stream_Video);
        Video_Count=1;
        Audio_Count=0;
        StreamKind.resize(0x100);
        StreamKind[Element_Name]=FromTS;
    }

    //New stream if needed
    int8u Stream_Pos=Element_Name-0xE0;
    if (Stream_Pos>=Video_Handle.size())
    {
        Video_Handle.resize(Stream_Pos+1);
        Video_TimeStamp.resize(Stream_Pos+1);
        Video_TimeStamp_End.resize(Stream_Pos+1);
    }
    if (Video_Handle[Stream_Pos]==NULL)
    {
        if (Element_Name==0xFD)
        {
            #if defined(MEDIAINFO_VC1_YES)
                Video_Handle[Stream_Pos]=new File_Vc1;
            #else
                //Filling
                Video_Handle[Stream_Pos]=new File__Base();
                Video_Handle[Stream_Pos]->Stream_Prepare(Stream_Video);
                Video_Handle[Stream_Pos]->Fill("Codec", "VC-1");
            #endif
        }
        else if (StreamKind.empty() || StreamKind[Element_Name]<=0x02)
        {
            #if defined(MEDIAINFO_MPEGV_YES)
                Video_Handle[Stream_Pos]=new File_Mpegv;
                ((File_Mpegv*)Video_Handle[Stream_Pos])->MPEG_Version=MPEG_Version;
            #else
                //Filling
                Video_Handle[Stream_Pos]=new File__Base();
                Video_Handle[Stream_Pos]->Stream_Prepare(Stream_Video);
                Video_Handle[Stream_Pos]->Fill("Codec", "MPEG Video");
            #endif
        }
        else if (StreamKind[Element_Name]==0x10)
        {
            #if defined(MEDIAINFO_MPEG4V_YES)
                Video_Handle[Stream_Pos]=new File_Mpeg4v;
            #else
                //Filling
                Video_Handle[Stream_Pos]=new File__Base();
                Video_Handle[Stream_Pos]->Stream_Prepare(Stream_Video);
                Video_Handle[Stream_Pos]->Fill("Codec", "MPEG-4V");
            #endif
        }
        else if (StreamKind[Element_Name]==0x1B)
        {
            #if defined(MEDIAINFO_AVC_YES)
                Video_Handle[Stream_Pos]=new File_Avc;
            #else
                //Filling
                Video_Handle[Stream_Pos]=new File__Base();
                Video_Handle[Stream_Pos]->Stream_Prepare(Stream_Video);
                Video_Handle[Stream_Pos]->Fill("Codec", "AVC");
            #endif
        }
        else
        {
            Video_Handle[Stream_Pos]=new File__Base();
            Video_Handle[Stream_Pos]->Stream_Prepare(Stream_Video);
            Video_Handle[Stream_Pos]->Fill("Codec", "Unsupported");
        }

        Video_TimeStamp[Stream_Pos]=TimeStamp;
    }

    //TimeStamp
    Video_TimeStamp_End[Stream_Pos]=TimeStamp;

    //Open Mpegv
    Open_Buffer_Init(Video_Handle[Stream_Pos], File_Size, File_Offset+Buffer_Offset);
    Open_Buffer_Continue(Video_Handle[Stream_Pos], Buffer+Buffer_Offset, Element_Size);

    //Testing other parsers in case of need
    if (StreamKind.empty() && Video_Handle[Stream_Pos]->Count_Get(Stream_General)==0)
    {
        FLUSH();
        ELEMENT(1, "Testing AVC...", Element_Size);
        if (Stream_Pos>=VideoAvc_Handle.size())
            VideoAvc_Handle.resize(Stream_Pos+1);
        if (VideoAvc_Handle[Stream_Pos]==NULL)
            VideoAvc_Handle[Stream_Pos]=new File_Avc;
        Open_Buffer_Init(VideoAvc_Handle[Stream_Pos], File_Size, File_Offset+Buffer_Offset);
        Open_Buffer_Continue(VideoAvc_Handle[Stream_Pos], Buffer+Buffer_Offset, Element_Size);

        if (VideoAvc_Handle[Stream_Pos]->Count_Get(Stream_General)>0)
        {
            ELEMENT(1, "AVC found, changing default parser", 0);
            delete Video_Handle[Stream_Pos]; Video_Handle[Stream_Pos]=VideoAvc_Handle[Stream_Pos]; VideoAvc_Handle[Stream_Pos]=NULL;
        }
        else
        {
            FLUSH();
            ELEMENT(1, "Testing MPEG-4 Video...", Element_Size);
            if (Stream_Pos>=VideoMpeg4v_Handle.size())
                VideoMpeg4v_Handle.resize(Stream_Pos+1);
            if (VideoMpeg4v_Handle[Stream_Pos]==NULL)
                VideoMpeg4v_Handle[Stream_Pos]=new File_Mpeg4v;
            Open_Buffer_Init(VideoMpeg4v_Handle[Stream_Pos], File_Size, File_Offset+Buffer_Offset);
            Open_Buffer_Continue(VideoMpeg4v_Handle[Stream_Pos], Buffer+Buffer_Offset, Element_Size);

            if (VideoAvc_Handle[Stream_Pos]->Count_Get(Stream_General)>0)
            {
                ELEMENT(1, "MPEG-4 Video found, changing default parser", 0);
                delete Video_Handle[Stream_Pos]; Video_Handle[Stream_Pos]=VideoMpeg4v_Handle[Stream_Pos]; VideoMpeg4v_Handle[Stream_Pos]=NULL;
            }
        }
    }

    if (Video_Count>0 && Video_Handle[Stream_Pos]->File_GoTo!=(int64u)-1)
        Video_Count--;

    //Demux
    DEMUX(Buffer+Buffer_Offset, Element_Size, Ztring::ToZtring(Element_Name, 16)+_T(".mpv"));
}

//---------------------------------------------------------------------------
void File_MpegPs::audioHD_stream()
{
    NAME ("Audio HD");

    //Demux
    DEMUX(Buffer+Buffer_Offset, Element_Size, Ztring::ToZtring(Element_Name, 16)+_T(".xxx"));
}

//***************************************************************************
// Helpers
//***************************************************************************

//---------------------------------------------------------------------------
bool File_MpegPs::Detect_NonMPEGPS ()
{
    //Element_Size
    if (Buffer_Size<8)
    {
        File_Offset=File_Size;
        return true;
    }

    //Detect DAT files, and the parser is not enough precise to detect them later
    if (CC4(Buffer)==CC4("RIFF"))
    {
        File_Offset=File_Size;
        return true;
    }

    //Detect TS files, and the parser is not enough precise to detect them later
    if (Buffer_Size>=188*4)
    {
        //Look for first Sync word
        while (Buffer_Offset<188 && CC1(Buffer+Buffer_Offset)!=0x47)
            Buffer_Offset++;
        if (Buffer_Offset<188 && CC1(Buffer+Buffer_Offset+188)==0x47 && CC1(Buffer+Buffer_Offset+188*2)==0x47 && CC1(Buffer+Buffer_Offset+188*3)==0x47)
        {
            File_Offset=File_Size;
            return true;
        }
        Buffer_Offset=0;
    }

    //Detect MPEG Version
    if (MPEG_Version==0)
        Detect_MPEG_Version();

    //Seems OK
    return false;
}

//---------------------------------------------------------------------------
void File_MpegPs::Detect_MPEG_Version ()
{
    /*
    size_t B5_Offset=Buffer_Offset;
    while (B5_Offset+4<=Buffer_Size && !(CC3(Buffer+B5_Offset)==0x000001 && CC1(Buffer+B5_Offset+3)==0xB5))
        B5_Offset++;
    if (B5_Offset+4>Buffer_Size)
        MPEG_Version=1;
    else
    */ //Currently default is MPEG-2, if Header says MPEG-1, this will be MPEG-1, no more.
        MPEG_Version=2;
}

//---------------------------------------------------------------------------
bool File_MpegPs::Detect_EOF_Needed ()
{
    //Jumping to the end of the file
    if (File_Size>=File_Offset+Buffer_Size+4*1024*1024 && File_Size!=(int64u)-1 && (
       (File_Offset>=2*1024*1024)
    //|| (Synched && Video_Count==0 && Audio_Count==0) //Subtitles were forgotten
    ))
    {
        //Details
        if (Config.Details_Get())
        {
            Details_Add_Error("-------------------------------------------");
            Details_Add_Error("---   MPEG-PS, Jumping to end of file   ---");
            Details_Add_Error("-------------------------------------------");
        }

        //Jumping
        File_GoTo=File_Size-4*1024*1024;
        return true;
    }
    return false;
}

//---------------------------------------------------------------------------
bool File_MpegPs::NextFrame()
{
    //Look for first Sync word
    if (Buffer_Offset_Temp==0) //Buffer_Offset_Temp is not 0 if NextFrame() has already parsed first frames
        Buffer_Offset_Temp=Buffer_Offset;
    while (Buffer_Offset_Temp+4<=Buffer_Size && !(CC3(Buffer+Buffer_Offset_Temp)==0x000001 && CC1(Buffer+Buffer_Offset_Temp+3)>=0xB9)) //0xB9 is the lower value for MPEG-PS
        Buffer_Offset_Temp++;

    //Not synched case
    if (!Synched && Buffer_Offset_Temp+4>Buffer_Size)
    {
        Buffer_Offset_Temp++;
        if (CC3(Buffer+Buffer_Offset_Temp)!=0x000001)
        {
            Buffer_Offset_Temp++;
            if (CC2(Buffer+Buffer_Offset_Temp)!=0x0000)
            {
                Buffer_Offset_Temp++;
                if (CC1(Buffer+Buffer_Offset_Temp)!=0x00)
                    Buffer_Offset_Temp++;
            }
        }
        Buffer_Offset=Buffer_Offset_Temp;
        Buffer_Offset_Temp=0;
        return false;
    }

    //Must wait more data?
    if (Buffer_Offset_Temp+4>Buffer_Size)
        return false;

    //Error in stream?
    if (Buffer_Offset_Temp-Buffer_Offset>0)
    {
        //There may have unuseful 0x00 in the stream
        if (Synched)
        {
            size_t Check=0;
            for (size_t Pos=Buffer_Offset; Pos<=Buffer_Offset_Temp; Pos++)
                Check+=Buffer[Pos];
            if (Check==0)
                Synched=false; //all zeros, this is synch
        }

        if (Synched)
            TRUSTED_ISNOT("Sync error", 0, Buffer_Offset_Temp-Buffer_Offset);
        else
        {
            ELEMENT(0, "Synchronization", Buffer_Offset_Temp-Buffer_Offset);
            FLUSH();
        }
    }

    //OK, we continue
    Buffer_Offset=Buffer_Offset_Temp;
    Buffer_Offset_Temp=0;
    Synched=true;
    return true;
}

//---------------------------------------------------------------------------
bool File_MpegPs::Element_Name_Get()
{
    //Enough data?
    if (Buffer_Offset+4>Buffer_Size)
        return false;

    //Element name
    Element_Name=BigEndian2int8u(Buffer+Buffer_Offset+3);

    return true;
}

//---------------------------------------------------------------------------
bool File_MpegPs::Element_Size_Get()
{
    switch (Element_Name)
    {
        //Fixed size
        case 0xB9 : //MPEG_program_end
            Element_HeaderSize=4; //MPEG start code + Element name
            Element_Size=0;
            Element_Next=Buffer_Offset+4; //MPEG start code + Element name
            break;

        //Not possible to know the size directly
        case 0xBA : //pack_start
            //Searching Element_Next
            Element_Next=Buffer_Offset+4;
            if (!Element_Size_Get_Element_Next())
                return false;
            //Filling
            Element_HeaderSize=4; //MPEG start code + Element name
            Element_Size=Element_Next-Buffer_Offset-4;
            break;

        //Header is only Size
        case 0xBB : //system_header_start
        case 0xBC : //program_stream_map
        case 0xBE : //padding_stream
        case 0xBF : //private_stream_2
        case 0xF0 : //ECM
        case 0xF1 : //EMM
        case 0xF2 : //DSMCC stream
        case 0xF8 : //ITU-T Rec. H .222.1 type E
        case 0xFF : //Program stream directory
            //Enough data?
            if (Buffer_Offset+6>Buffer_Size)
                return false;
            //Filling
            Element_HeaderSize=6; //MPEG start code + Element name + Element size
            Element_Size=BigEndian2int16u(Buffer+Buffer_Offset+4);
            Element_Next=Buffer_Offset+6+Element_Size; //MPEG start code + Element name + Element size
            break;

        //Element with PES Header (with Size in Header)
        default :
            if (!Buffer_Parse_Header()) //Header
                return false;
    }

    //Testing
    if (Element_Size==0) //Not normal
        TRUSTED_ISNOT("Next element is too small, not normal");
    if (Element_Next+2<=Buffer_Size && CC2(Buffer+Element_Next)!=0x0000)
        TRUSTED_ISNOT("Next Synch is not found");

    return true;
}

//---------------------------------------------------------------------------
bool File_MpegPs::Element_Size_Get_Unlimited()
{
    //Searching Element_Next
    Element_Next=Buffer_Offset;
    if (!Element_Size_Get_Element_Next())
        Element_Next=Buffer_Size; //The whole buffer

    Element_HeaderSize=0;
    Element_Size=Element_Next-Buffer_Offset;
    Synched=true;
    return true;
}

//---------------------------------------------------------------------------
bool File_MpegPs::Element_Size_Get_Element_Next()
{
    //Integrity
    if (Element_Next<Buffer_Offset)
        Element_Next=Buffer_Size;
    if (Element_Next==Buffer_Size && CC3(Buffer+Element_Next)==0x000001 && CC1(Buffer+Element_Next+3)>=0xB9) //0xB9 is the lower value for MPEG-PS
        Element_Next+=4; //Not normal, risk of infinite loop

    //Searching Element_Next
    while (Element_Next+4<=Buffer_Size && !(CC3(Buffer+Element_Next)==0x000001 && CC1(Buffer+Element_Next+3)>=0xB9)) //0xB9 is the lower value for MPEG-PS
        Element_Next++;
    if (Element_Next+4>Buffer_Size)
    {
        if (File_Offset+Element_Next+4<=File_Size)
            return false; //This is not the end of file, so we must wait for more data

        //Managing special cases
        Element_Next=Buffer_Size; //End of file
    }

    return true;
}

//---------------------------------------------------------------------------
bool File_MpegPs::Element_Name_IsOK()
{
    //Is it the start of a MPEG PS?
    if (Count_Get(Stream_General)==0
     &&   Element_Name!=0xBA
     && !(Element_Name>=0xBD && Element_Name<=0xEF)
     &&   Element_Name!=0xFD //HD-DVD
     )
    {
        //MPEG PS didn't begin, skipping
        //Details
        if (Config.Details_Get()>0.9)
        {
            Details_Add_Element(0, Element_Name, Element_Size-4);
            Details_Add_Element("Has not began, skipping");
        }
        Buffer_Offset+=Element_HeaderSize+Element_Size;
        return false;
    }

    return true;
}

//---------------------------------------------------------------------------
// Packet header
bool File_MpegPs::Buffer_Parse_Header()
{
    //There is a PES Header with size and long packet header
    if (Buffer_Offset+6>=Buffer_Size)
        return false; //No enough bytes

    //Element size
    Element_Size=BigEndian2int16u(Buffer+Buffer_Offset+4); //Element size
    Element_HeaderSize=6;

    //Element_Next
    if (Element_Size==0)
    {
        //Video stream in Transport Stream, Element_Size may be unlimited
        Video_Unlimited=true;
        Element_Next=Buffer_Offset+6; ////MPEG start code + Element name + Element size
        if (!Element_Size_Get_Element_Next())
            Element_Next=Buffer_Size; //The whole buffer
        if (Element_Next>Buffer_Offset && Element_Next-Buffer_Offset>Element_HeaderSize)
            Element_Size=Element_Next-Buffer_Offset-Element_HeaderSize;
        else
            Element_Size=0; //There is a problem...
    }
    else
    {
        Element_Next=Buffer_Offset+6+Element_Size;

        //Testing end of the file before the end of the PES
        if (File_Offset+Element_Next>=File_Size) //This is the end of the file
            Element_Next=File_Size;

        //Testing if we have enough data
        if (Element_Next>Buffer_Size)
            return false;
    }

    //Parsing header data
    if (!Buffer_Parse_Header_Data()) //Header
        return false;

    return true;
}

//---------------------------------------------------------------------------
// Packet header data
bool File_MpegPs::Buffer_Parse_Header_Data()
{
    TimeStamp.Have_PTS=false;
    TimeStamp.Have_DTS=false;

    //MPEG-2
    switch (MPEG_Version)
    {
        case 1  : if (!Buffer_Parse_Header_Data_MPEG1()) return false; break;
        case 2  : if (!Buffer_Parse_Header_Data_MPEG2()) return false; break;
        default : return true; //Change nothing
    }

    Element_Size=Element_Next-Buffer_Offset-Element_HeaderSize;
    return true;
}

//---------------------------------------------------------------------------
// Packet header data - MPEG-1
bool File_MpegPs::Buffer_Parse_Header_Data_MPEG1()
{
    //Keep out firsts 0xFF
    while (Buffer_Offset+Element_HeaderSize<Element_Next && Buffer[Buffer_Offset+Element_HeaderSize]==0xFF)
        Element_HeaderSize++;
    if (Buffer_Offset+Element_HeaderSize>=Element_Next)
        return true;

    //Buffer scale and size
    if ((Buffer[Buffer_Offset+Element_HeaderSize]&0xC0)==0x40)
    {
        Element_HeaderSize+=2;
    }

    //With PTS only
    if ((Buffer[Buffer_Offset+Element_HeaderSize]&0xF0)==0x20)
    {
        if (Element_Size<5)
            return false;
        BitStream BS(Buffer+Buffer_Offset+Element_HeaderSize, Element_Size);
        TimeStamp.Have_PTS=true;
        TimeStamp.Have_DTS=false;
        TimeStamp.PTS=MPEG_TimeStamp_Read(BS);
        Element_HeaderSize+=5;
    }

    //With PTS and DTS
    else if ((Buffer[Buffer_Offset+Element_HeaderSize]&0xF0)==0x30)
    {
        if (Element_Size<10)
            return false;
        BitStream BS(Buffer+Buffer_Offset+Element_HeaderSize, Element_Size);
        TimeStamp.Have_PTS=true;
        TimeStamp.Have_DTS=true;
        TimeStamp.PTS=MPEG_TimeStamp_Read(BS);
        Element_HeaderSize+=10;
    }

    return true;
}

//---------------------------------------------------------------------------
// Packet header data - MPEG-2
bool File_MpegPs::Buffer_Parse_Header_Data_MPEG2()
{
    if (Element_Size<3)
        return false;

    //Reading
    BitStream BS(Buffer+Buffer_Offset+6, Element_Size-6);
    int32u Flags, Size;
    BS.Skip(2);                                 //01
    BS.Skip(2);                                 //PES_scrambling_control
    BS.Skip(1);                                 //PES_priority
    BS.Skip(1);                                 //data_alignment_indicator
    BS.Skip(1);                                 //copyright
    BS.Skip(1);                                 //original_or_copy
    Flags=BS.Get(2);                            //PTS_DTS_flags
    BS.Skip(1);                                 //ESCR_flag
    BS.Skip(1);                                 //ES_rate_flag
    BS.Skip(1);                                 //DSM_trick_mode_flag
    BS.Skip(1);                                 //additional_copy_info_flag
    BS.Skip(1);                                 //PES_CRC_flag
    BS.Skip(1);                                 //PES_extension_flag
    Size=BS.Get(8);                             //PES_header_data_length

    Element_HeaderSize+=3;
    if (3+Size>Element_Size)
        return false;

    //Options
    if (Flags==0x2)
    {
        TimeStamp.Have_PTS=true;
        TimeStamp.Have_DTS=false;
        TimeStamp.PTS=MPEG_TimeStamp_Read(BS);
    }
    if (Flags==0x3)
    {
        TimeStamp.Have_PTS=true;
        TimeStamp.Have_DTS=true;
        TimeStamp.PTS=MPEG_TimeStamp_Read(BS);
        TimeStamp.DTS=MPEG_TimeStamp_Read(BS);
    }

    //Positioning Packet data start
    Element_HeaderSize+=Size;
    return true;
}

//---------------------------------------------------------------------------
int64u File_MpegPs::MPEG_TimeStamp_Read (BitStream &BS)
{
    int64u TimeStamp;

    BS.Skip(4);
    TimeStamp=BS.Get(3);
    BS.Skip(1);
    TimeStamp<<=15;
    TimeStamp|=BS.Get(15);
    BS.Skip(1);
    TimeStamp<<=15;
    TimeStamp|=BS.Get(15);
    BS.Skip(1);
    return TimeStamp;
}

//***************************************************************************
// Information
//***************************************************************************

//---------------------------------------------------------------------------
void File_MpegPs::HowTo(stream_t StreamKind)
{
    switch (StreamKind)
    {
        case (Stream_General) :
            Fill_HowTo("Format", "R");
            Fill_HowTo("OveralBitRate", "R");
            Fill_HowTo("PlayTime", "R");
            Fill_HowTo("Encoded_Application", "R");
            break;
        case (Stream_Video) :
            Fill_HowTo("Codec", "R");
            Fill_HowTo("BitRate", "R");
            Fill_HowTo("Width", "R");
            Fill_HowTo("Height", "R");
            Fill_HowTo("AspectRatio", "R");
            Fill_HowTo("FrameRate", "R");
            Fill_HowTo("Delay", "R");
            break;
        case (Stream_Audio) :
            Fill_HowTo("Codec", "R");
            Fill_HowTo("BitRate", "R");
            Fill_HowTo("Channel(s)", "R");
            Fill_HowTo("SamplingRate", "R");
            Fill_HowTo("Resolution", "R");
            Fill_HowTo("Delay", "R");
            break;
        case (Stream_Text) :
            Fill_HowTo("Codec", "R");
            Fill_HowTo("Delay", "R");
            break;
        case (Stream_Chapters) :
            break;
        case (Stream_Image) :
            break;
        case (Stream_Menu) :
            break;
        case (Stream_Max) :
            break;
    }
}

} //NameSpace

#endif //MEDIAINFO_MPEGPS_YES

