// File_Mpegts - Info for MPEG Transport Stream files
// Copyright (C) 2006-2007 Jerome Martinez, Zen@MediaArea.net
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
//
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

//---------------------------------------------------------------------------
// Compilation conditions
#include <MediaInfo/Setup.h>
#ifdef __BORLANDC__
    #pragma hdrstop
#endif
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#if defined(MEDIAINFO_MPEGTS_YES)
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
#include "MediaInfo/Multiple/File_MpegTs_Pms.h"
#include "ZenLib/BitStream.h"
using namespace ZenLib;
//---------------------------------------------------------------------------

namespace MediaInfoLib
{

//***************************************************************************
// Constants
//***************************************************************************

//---------------------------------------------------------------------------
const char* MpegTs_Pms_StreamID(int32u ID)
{
    switch (ID)
    {
        case 0x00 : return "ITU-T | ISO/IEC Reserved";
        case 0x01 : return "ISO/IEC 11172 Video";
        case 0x02 : return "ITU-T Rec. H.262 | ISO/IEC 13818-2 Video or ISO/IEC 11172-2 constrained parameter video stream";
        case 0x03 : return "ISO/IEC 11172 Audio";
        case 0x04 : return "ISO/IEC 13818-3 Audio";
        case 0x05 : return "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 private_sections";
        case 0x06 : return "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 PES packets containing private data";
        case 0x07 : return "ISO/IEC 13522 MHEG";
        case 0x08 : return "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 Annex A DSM-CC";
        case 0x09 : return "ITU-T Rec. H.222.1";
        case 0x0A : return "ISO/IEC 13818-6 type A";
        case 0x0B : return "ISO/IEC 13818-6 type B";
        case 0x0C : return "ISO/IEC 13818-6 type C";
        case 0x0D : return "ISO/IEC 13818-6 type D";
        case 0x0E : return "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 auxiliary";
        case 0x0F : return "ISO/IEC 13818-7 Audio with ADTS transport syntax";
        case 0x10 : return "ISO/IEC 14496-2 Visual";
        case 0x11 : return "ISO/IEC 14496-3 Audio with the LATM transport syntax as defined in ISO/IEC 14496-3 / AMD 1";
        case 0x12 : return "ISO/IEC 14496-1 SL-packetized stream or FlexMux stream carried in PES packets";
        case 0x13 : return "ISO/IEC 14496-1 SL-packetized stream or FlexMux stream carried in ISO/IEC14496_sections.";
        case 0x14 : return "ISO/IEC 13818-6 Synchronized Download Protocol";
        case 0x1B : return "H.264";
        default :
            if (ID>=0x15
             && ID<=0x7F) return "ITU-T Rec. H.222.0 | ISO/IEC 13818-1 reserved";
            if (ID>=0x80
             && ID<=0xFF) return "User Private";
            return "forbidden";
    }
}

//---------------------------------------------------------------------------
const char* MpegTs_Pms_audio_type(int8u ID)
{
    switch (ID)
    {
        case 0x00 : return "Undefined";
        case 0x01 : return "Clean effects";
        case 0x02 : return "Hearing impaired";
        case 0x03 : return "Visual impaired commentary";
        default :   return "Rreserved";
    }
}

//***************************************************************************
// Format
//***************************************************************************

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Read_Buffer_Init()
{
    section_length=0;
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Read_Buffer_Continue()
{
    //Header
    if (section_length==0)
    {
        Header();
        if (section_length==0)
            FINNISHED(); //Problem...
    }


    //Loop
    while (Buffer_Parse());

    //Positionning
    section_length-=Buffer_Offset;
}

//***************************************************************************
// Buffer
//***************************************************************************

//---------------------------------------------------------------------------
bool File_MpegTs_Pms::Buffer_Parse()
{
    //Parsing
    if (!Element_Parse())
        return false;

    //Is it the end?
    if (Buffer_Offset+4>=4+section_length)
    {
        Footer();
        FINNISHED_BOOL();
    }

    return true;
}

//---------------------------------------------------------------------------
bool File_MpegTs_Pms::Element_Parse()
{
    FLUSH();

    //Configure defines : must return false
    #undef INTEGRITY_RETURNVALUE
    #define INTEGRITY_RETURNVALUE false

    //Parsing
    Element_Size=5;
    int32u stream_type, elementary_PID;
    BS_BEGIN();
    GET_BS ( 8, stream_type,                                stream_type); INFO(MpegTs_Pms_StreamID(stream_type));
    SKIP_BS( 3,                                             reserved);
    GET_BS (13, elementary_PID,                             elementary_PID);
    SKIP_BS( 4,                                             reserved);
    GET_BS (12, stream_info_length,                         ES_info_length);
    BS_END();

    //UnConfigure defines : must return nothing
    #undef INTEGRITY_RETURNVALUE
    #define INTEGRITY_RETURNVALUE

    if (Buffer_Offset+5+stream_info_length>Buffer_Size)
        return false;

    ELEMENT(0, (int16u)elementary_PID, 5+stream_info_length); NAME(MpegTs_Pms_StreamID(stream_type));

    //Filling
    Stream_Prepare(Stream_General);
    Fill("ID", elementary_PID);
    Fill("UniqueID", stream_type); //stream_type is in UniqueID only for optimisation (standart tags)

    //Descriptors
    Buffer_Offset+=BS.Offset_Get();
    Descriptors();
    Buffer_Offset+=stream_info_length;

    FLUSH();
    return true;
}

//***************************************************************************
// Elements
//***************************************************************************

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Header()
{
    ELEMENT(0, "Header", 13);

    //Parsing
    Element_Size=13;
    BS_BEGIN();
    SKIP_BS( 8,                                                 pointer_field);
    SKIP_BS( 8,                                                 table_id);
    SKIP_BS( 1,                                                 section_syntax_indicator);
    SKIP_BS( 1,                                                 zero);
    SKIP_BS( 2,                                                 reserved);
    GET_BS (12, section_length,                                 section_length);
    if (section_length<9)
    {
        section_length=0;
        return;
    }
    SKIP_BS(16,                                                 program_number);
    SKIP_BS( 2,                                                 reserved);
    SKIP_BS( 5,                                                 version_number);
    SKIP_BS( 1,                                                 current_next_indicator);
    SKIP_BS( 8,                                                 section_number);
    SKIP_BS( 8,                                                 last_section_number);
    SKIP_BS( 3,                                                 reserved);
    SKIP_BS(13,                                                 PCR_PID);
    SKIP_BS( 4,                                                 reserved);
    GET_BS (12, stream_info_length,                             program_info_length);

    //Enough data?
    if (13+stream_info_length>Buffer_Size)
    {
        section_length=0; //We want to redo it with more data
        return;
    }

    //Descriptors and Positionning
    Buffer_Offset+=BS.Offset_Get();
    Descriptors();
    Buffer_Offset+=stream_info_length;

    FLUSH();
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Footer()
{
    ELEMENT(0, "Footer", Buffer_Size-Buffer_Offset);

    //Parsing
    Element_Size=Buffer_Size-Buffer_Offset;
    BS_BEGIN();
    SKIP_BS(32,                                                 CRC_32);
    BS_END_FF();

    FLUSH();
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Descriptors()
{
    FLUSH();
    ELEMENT_BEGIN();
    ELEMENT("Descriptors", stream_info_length);

    //Parsing
    while (stream_info_length>0)
        Descriptor();

    ELEMENT_END();
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Descriptor()
{
    FLUSH();
    ELEMENT_BEGIN();

    //Parsing
    Element_Size=2;
    int32u descriptor_tag, descriptor_length;
    CC_BEGIN();
    GET_B1 (descriptor_tag,                                     descriptor_tag);
    GET_B1 (descriptor_length,                                  descriptor_length);
    CC_END();

    ELEMENT(Ztring().From_CC2(descriptor_tag), descriptor_length);

    //Positioning
    Buffer_Offset+=2;
    stream_info_length-=2;
    Element_Size=descriptor_length;

    #define ELEMENT_CASE(_NAME, _DETAIL) \
        case 0x##_NAME : NAME(_DETAIL); Descriptor_##_NAME(); break;


    switch (descriptor_tag)
    {
        ELEMENT_CASE(00, "Reserved");
        ELEMENT_CASE(01, "Reserved");
        ELEMENT_CASE(02, "video_stream_descriptor");
        ELEMENT_CASE(03, "audio_stream_descriptor");
        ELEMENT_CASE(04, "hierarchy_descriptor");
        ELEMENT_CASE(05, "registration_descriptor");
        ELEMENT_CASE(06, "data_stream_alignment_descriptor");
        ELEMENT_CASE(07, "target_background_grid_descriptor");
        ELEMENT_CASE(08, "Video_window_descriptor");
        ELEMENT_CASE(09, "CA_descriptor");
        ELEMENT_CASE(0A, "ISO_639_language_descriptor");
        ELEMENT_CASE(0B, "System_clock_descriptor");
        ELEMENT_CASE(0C, "Multiplex_buffer_utilization_descriptor");
        ELEMENT_CASE(0D, "Copyright_descriptor");
        ELEMENT_CASE(0E, "Maximum_bitrate_descriptor");
        ELEMENT_CASE(0F, "Private_data_indicator_descriptor");
        ELEMENT_CASE(10, "Smoothing_buffer_descriptor");
        ELEMENT_CASE(11, "STD_descriptor");
        ELEMENT_CASE(12, "IBP_descriptor");
        ELEMENT_CASE(13, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(14, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(15, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(16, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(17, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(18, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(19, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(1A, "Defined in ISO/IEC 13818-6");
        ELEMENT_CASE(1B, "MPEG-4_video_descriptor");
        ELEMENT_CASE(1C, "MPEG-4_audio_descriptor");
        ELEMENT_CASE(1D, "IOD_descriptor");
        ELEMENT_CASE(1E, "SL_descriptor");
        ELEMENT_CASE(1F, "FMC_descriptor");
        ELEMENT_CASE(20, "External_ES_ID_descriptor");
        ELEMENT_CASE(21, "MuxCode_descriptor");
        ELEMENT_CASE(22, "FmxBufferSize_descriptor");
        ELEMENT_CASE(23, "MultiplexBuffer_descriptor");
        default: if (descriptor_tag>=64)
                    NAME("user private")
                 else
                    NAME("unknown");
                 break;
    }

    ELEMENT_END();

    //Positioning
    Buffer_Offset+=descriptor_length;
    stream_info_length-=descriptor_length;
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Descriptor_0A()
{
    //Parsing
    int32u ISO_639_language_code, audio_type;
    CC_BEGIN();
    GET_C3 (ISO_639_language_code,                              ISO_639_language_code);
    GET_C1 (audio_type,                                         audio_type); INFO(MpegTs_Pms_audio_type(audio_type));
    CC_END();
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Descriptor_0E()
{
    //Parsing
    BS_BEGIN();
    SKIP_BS( 2,                                                 reserved);
    SKIP_BS(22,                                                 maximum_bitrate);
    BS_END();
}

//---------------------------------------------------------------------------
void File_MpegTs_Pms::Descriptor_0F()
{
    //Parsing
    CC_BEGIN();
    SKIP_B4(                                                    private_data_indicator);
    CC_END();
}

//***************************************************************************
// C++
//***************************************************************************

} //NameSpace

#endif //MEDIAINFO_MPEGTS_YES
